//The code is released under GPLv2 (http://www.gnu.org/licenses/gpl-2.0.html),
//by Laszlo Toth.
//Use the code at your own responsibility.
//For help and disclaimer please visit:
//  http://www.soonerorlater.hu/index.khtml?article_id=517

#include <string>
#include <openssl/evp.h> 
#include <openssl/sha.h>
#include <openssl/md5.h>
#include <iostream>
#include <boost/python/module.hpp>
#include <boost/python/def.hpp>


namespace{

	int HexStringtoBinArray(const char* str, unsigned char* array);

	std::string decrypt_sesskey(std::string sesskey, std::string hash, int keylen){
		unsigned char bin_sesskey[64];
		unsigned char bin_decr_sesskey[256];
		unsigned char bin_hash[32];
		unsigned char iv[32]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
		char tmp[3];
		std::string decr_sesskey;
		int outlen;
		int outlen2;
		int inlen;
		EVP_CIPHER_CTX ctx;

		inlen=sesskey.length()/2;
		
		memset(bin_hash,0,32);
		HexStringtoBinArray(sesskey.c_str(), bin_sesskey);
		HexStringtoBinArray(hash.c_str(), bin_hash);
		

		if(keylen==192){
			EVP_CIPHER_CTX_init(&ctx);
			EVP_DecryptInit_ex(&ctx, EVP_aes_192_cbc(), NULL, bin_hash, iv);
			EVP_DecryptUpdate(&ctx, bin_decr_sesskey, &outlen, bin_sesskey, inlen);
			EVP_DecryptFinal_ex(&ctx, bin_decr_sesskey+32, &outlen2);
			EVP_CIPHER_CTX_cleanup(&ctx);
		}else if(keylen==128){
			EVP_CIPHER_CTX_init(&ctx);
			EVP_DecryptInit_ex(&ctx, EVP_aes_128_cbc(), NULL, bin_hash, iv);
			EVP_DecryptUpdate(&ctx, bin_decr_sesskey, &outlen, bin_sesskey, inlen);
			EVP_DecryptFinal_ex(&ctx, bin_decr_sesskey+outlen, &outlen2);
			EVP_CIPHER_CTX_cleanup(&ctx);
		}
		
		memset(tmp,0,3);
		for(int i=0;i<48;i++){
			sprintf(tmp,"%02X", bin_decr_sesskey[i]);
			decr_sesskey=decr_sesskey+tmp;
		}
		return decr_sesskey;
	}

	std::string combine_skeys_11g(std::string decr_srv_skey, std::string decr_clnt_skey){
		unsigned char bin_decr_srv_skey[256];
		unsigned char bin_decr_clnt_skey[256];
		char tmp[3];
		unsigned char csk[24];
		unsigned char md1[16];
		unsigned char md2[16];
		std::string combined_skey;


		HexStringtoBinArray(decr_srv_skey.c_str(), bin_decr_srv_skey);
		HexStringtoBinArray(decr_clnt_skey.c_str(), bin_decr_clnt_skey);
	
		for(int m=0; m<24; m++){
			csk[m]=bin_decr_srv_skey[m+16] ^ bin_decr_clnt_skey[m+16];
		}
		MD5(csk, 16, md1);
		MD5(csk+16,8,md2);
		memcpy(csk,md1,16);
		memcpy(csk+16,md2,8);

		memset(tmp,0,3);
		for(int i=0;i<24;i++){
			sprintf(tmp,"%02X", csk[i]);
			combined_skey=combined_skey+tmp;
		}

		return combined_skey;

	}

	std::string combine_skeys_10g(std::string decr_srv_skey, std::string decr_clnt_skey){
			unsigned char bin_decr_srv_skey[256];
			unsigned char bin_decr_clnt_skey[256];
			char tmp[3];
			unsigned char csk[24];
			unsigned char md[16];
			std::string combined_skey;


			HexStringtoBinArray(decr_srv_skey.c_str(), bin_decr_srv_skey);
			HexStringtoBinArray(decr_clnt_skey.c_str(), bin_decr_clnt_skey);
			
			for(int m=0; m<16; m++){
				csk[m]=bin_decr_srv_skey[m+16] ^ bin_decr_clnt_skey[m+16];
			}
			MD5(csk, 16, md);

			memset(tmp,0,3);
			for(int i=0;i<16;i++){
				sprintf(tmp,"%02X", md[i]);
				combined_skey=combined_skey+tmp;
			}

			return combined_skey;

	}

	std::string decrypt_password(std::string auth_passwd, std::string csk, int keylen){
		unsigned char bin_auth_passwd[256];
		unsigned char bin_csk[256];
		unsigned char bin_passwd[256];
		unsigned char iv[32]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
		char tmp[3];
		std::string password;
		int outlen;
		int outlen2;
		int inlen;
		EVP_CIPHER_CTX ctx;

		inlen=auth_passwd.length()/2;
		
		memset(bin_passwd,0,32);
		HexStringtoBinArray(auth_passwd.c_str(), bin_auth_passwd);
		HexStringtoBinArray(csk.c_str(), bin_csk);
		

		if(keylen==192){
			EVP_CIPHER_CTX_init(&ctx);
			EVP_DecryptInit_ex(&ctx, EVP_aes_192_cbc(), NULL, bin_csk, iv);
			EVP_DecryptUpdate(&ctx, bin_passwd, &outlen, bin_auth_passwd, inlen);
			EVP_DecryptFinal_ex(&ctx, bin_passwd+outlen, &outlen2);
			EVP_CIPHER_CTX_cleanup(&ctx);
		}else if(keylen==128){
			EVP_CIPHER_CTX_init(&ctx);
			EVP_DecryptInit_ex(&ctx, EVP_aes_128_cbc(), NULL, bin_csk, iv);
			EVP_DecryptUpdate(&ctx, bin_passwd, &outlen, bin_auth_passwd, inlen);
			EVP_DecryptFinal_ex(&ctx, bin_passwd+outlen, &outlen2);
			EVP_CIPHER_CTX_cleanup(&ctx);
		}
		
		memset(tmp,0,3);
		for(int i=16;i<outlen+outlen2;i++){
			sprintf(tmp,"%c", bin_passwd[i]);
			password=password+tmp;
		}
		return password;
	}

	std::string encrypt_password(std::string password, std::string csk, int keylen){
		unsigned char bin_auth_passwd[256];
		unsigned char bin_csk[256];
		unsigned char bin_passwd[256];
		unsigned char iv[32]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
		char tmp[3];
		std::string auth_passwd;
		int outlen;
		int outlen2;
		int inlen;
		EVP_CIPHER_CTX ctx;

		inlen=password.length()/2;
		
		memset(bin_passwd,0,256);
		HexStringtoBinArray(password.c_str(), bin_passwd);
		HexStringtoBinArray(csk.c_str(), bin_csk);
		

		if(keylen==192){
			EVP_CIPHER_CTX_init(&ctx);
			EVP_EncryptInit_ex(&ctx, EVP_aes_192_cbc(), NULL, bin_csk, iv);
			EVP_EncryptUpdate(&ctx, bin_auth_passwd, &outlen, bin_passwd, inlen);
			EVP_EncryptFinal_ex(&ctx, bin_auth_passwd+outlen, &outlen2);
			EVP_CIPHER_CTX_cleanup(&ctx);
		}else if(keylen==128){
			EVP_CIPHER_CTX_init(&ctx);
			EVP_EncryptInit_ex(&ctx, EVP_aes_128_cbc(), NULL, bin_csk, iv);
			EVP_EncryptUpdate(&ctx, bin_auth_passwd, &outlen, bin_passwd, inlen);
			EVP_EncryptFinal_ex(&ctx, bin_auth_passwd+outlen, &outlen2);
			EVP_CIPHER_CTX_cleanup(&ctx);
		}
		
		memset(tmp,0,3);
		for(int i=0;i<outlen+outlen2;i++){
			sprintf(tmp,"%02X", bin_auth_passwd[i]);
			auth_passwd=auth_passwd+tmp;
		}
		return auth_passwd;
	}

	

	int HexStringtoBinArray(const char* str, unsigned char* array){
	
		unsigned char t[2];
		unsigned int hexc;
		int j=0;

		for(int i=0;i<strlen(str);i=i+2){
			t[0]=str[i];
			t[1]=str[i+1];
			hexc = t[0]-48;
			if (hexc > 9) hexc-=7;
			array[j]=hexc*16;
			hexc = t[1]-48;
			if (hexc > 9) hexc-=7;
			array[j]+=hexc;
			j++;
		}
		return j;
	}
}


namespace python = boost::python;

BOOST_PYTHON_MODULE(aesdecrypt)
{
    // Add regular functions to the module.
    python::def("decrypt_sesskey", decrypt_sesskey);
    python::def("combine_skeys_11g", combine_skeys_11g);
    python::def("combine_skeys_10g", combine_skeys_10g);
    python::def("decrypt_password", decrypt_password);
    python::def("encrypt_password", encrypt_password);
}




